/*
!
! KROC ALPHA assembly-language scheduler.
!
! $Source: /extra/play/cvs/kroc/src/target/alpha/sched.s,v $
!
! $Id: sched.s,v 1.1 1999/09/27 21:10:12 rivimey Exp $
!
! (C) Copyright 1995 M.D. Poole <M.D.Poole@ukc.ac.uk>
!                    et al.
! University of Kent at Canterbury
!
! This file contains functions that emulate the Transputer hardware scheduler,
! and code for allocating workspace, initializing registers, and entering the
! user's program.
!
! Original Sparc version written by:
!      Kneale J. Rothwell, Graham Shaw, and Keith E. Woollacott
! Optimization, endian swapping, etc. by:
!      David C. Wood
! Hard channels and timer queue stuff by:
!      Peter H. Welch and David C. Wood
! Recoded for DEC Alpha by M. D. Poole starting on 17th March 1995
!
*/
/*{{{  notes about defines that produce different versions*/
/*
HOST_OS_OSF30 is used to flag Unix signal implementation of kbd, time
KBD_SIGNAL    says keys are signalled
TIM_SIGNAL    says timeouts are signalled

_3L           is set if compiling with 3L C on an alphadata board
HARDLINKS     is defined for (threaded) external communications on alphadata
ALONE         is set for non-root processors with no host access
*/
/*}}}  */
/*{{{  notes on this particular version*/
/* This version either calls rtime or uses an rpcc instruction to do timing
   according to whether HOST_OS_OSF30 is defined. The clock will tick
   respectively at 1 per microsecond, or 235 per microsecond (observed on
   aztec workstation NOT universally true */
/* This version uses s4 for Wptr MDP 7/2/96 */
/* This version requires the following support from code written in C,
   or in library:

   __divqu     # assume does not corrupt t0,t1,t2
   __remqu     # assume does not corrupt t0,t1,t2
   __divl
   __reml
   sqrt
   fputc
   fflush

   read_the_kbd
   acknowledge
   setalarm
   safe_pause
   dead
   occam_exit

   _fpdata
   kroc_err
   kroc_out
   rtime
   ualarm
   impossible
*/
/*}}}  */

#ifdef _3L
#define HARDLINKS
#endif
#include <regdef.h>
#include "sched.h"
/*{{{  Unix signals supported?*/
#if defined(HOST_OS_OSF30)
#define KBD_SIGNAL
#define TIM_SIGNAL
#else
#endif
/*}}}  */

	.set noreorder
	.set noat

/*{{{  floating regs*/
#define F0      $f0
#define F1      $f1
#define F2      $f2
#define F3      $f3
#define F4      $f4
#define F5      $f5
#define F6      $f6
#define F7      $f7
#define F8      $f8
#define F9      $f9
#define F10      $f10
#define F11      $f11
#define F12      $f12
#define F13      $f13
#define F14      $f14
#define F15      $f15
#define F16      $f16
#define F17      $f17
#define F18      $f18
#define F19      $f19
#define F20      $f20
#define F21      $f21
#define F22      $f22
#define F23      $f23
#define F24      $f24
#define F25      $f25
#define F26      $f26
#define F27      $f27
#define F28      $f28
#define F29      $f29
#define F30      $f30
#define F31      $f31
#define FA       F0
/*}}}  */

/*{{{  notes on register usage*/
/*
   Occam workspace is allocated above program code, and vector space above that.

   All entries are assumed to be called with return link saved in L

   When a routine here is emulating a transputer instruction the following
   conventions usually apply, and if so it is assumed that code compiled
   from occam will use registers in conformance with them.
   This may involve explicit register moves before and/or after the call.

   A single parameter (Transputer Areg) will be found in REG_0.
   Two parameters (Transputer Breg, Areg) will be found in REG_0, REG_1
   Three parameters (Transputer Creg, Breg, Areg) are in REG_0,REG_1,REG_2

   Results are returned similarly.

   The assembler can generate uses of the registers t9, t10, t11 which
   we call T1, T2, T3 - and so these are used for very local purposes only.

   The register P5 is used as an additional local temporary exclusively
   for the boolean results of comparisons.

   C functions are called with zero, one or two arguments using P0, P1
   C function results are expected to be found in R0

   Some routines here obey different conventions:
   These are those like $div and $rem that emulate extended instructions
   sometimes by calling routines supplied by the C runtime system.

   Arguments for these routines are passed in in t10, t11 and results returned
   in t12.

   The constant MOSTNEG_INT32 (MIN) is held permanently in the register s0
   The registers s1, s2, s3 are used for Queue pointers Front, Back, TimeQ
   Register s4 represents the Transputer Wptr in all occam code.
   Register s5 (L) is used as return address to occam code.
   Register s6 (fp) must always hold the address of occam_table while
   occam code is running - as C functions must preserve and restore it, it
   only needs to be set up at entry to occam.

   References to local data and to external C routines implicitly use
   the gp register.
*/
/*}}}  */

/*{{{  compile time constants*/
W_0               = Temp              #  Process workspace offsets renamed
W_Iptr            = Iptr
W_Link            = Link
W_State           = State
W_Pointer         = W_State
W_TLink           = TLink
W_Time            = Time

Local1          = 1*4                   #  Extra locals for hard channels
Local2          = 2*4

SIZE_FILE       = 56                    #  sizeof(FILE)
/*}}}  */
	.data
/*{{{  variables and other data*/
	.globl variables
	.globl kbd_sync
#ifdef HARDLINKS
	.globl link_sync
#endif
	.globl tim_sync
	.globl SaveFP
	.globl OccErrPtr
	.globl kbd_chan
	.globl scr_chan
	.globl err_chan
variables:
kbd_sync:
	.long   0
#ifdef HARDLINKS
link_sync:
	.long   0
#endif
tim_sync:
	.long   0
mostnegval:
	.long   MostNeg
OccErrPtr:
	.quad   0
SaveFP:
	.quad   0
	.quad   0
Savesp:
	.quad   0
#ifndef ALONE
kbd_chan:
	.long   MostNeg                 #  NotProcess.p
scr_chan:
	.long   scr_workspace
err_chan:
	.long   err_workspace
!       Workspace for keyboard process (stdin)

	.long   MostNeg                 #  W_State/W_Pointer
	.long   MostNeg                 #  W_Link
	.long   MostNeg                 #  W_Iptr
kbd_workspace:
	.long   MostNeg                 #  W_0

!       Workspace for screen process (stdout)

	.long   scr_workspace           #  W_Pointer
	.long   MostNeg                 #  W_Link
	.long   hard_print              #  W_Iptr
scr_workspace:
	.long   0                       #  W_0
	.long   scr_chan                #  Local 1
	.long   0                       #  Local 2

!       Workspace for error process (stderr)

	.long   err_workspace           #  W_Pointer
	.long   MostNeg                 #  W_Link
	.long   hard_print              #  W_Iptr
err_workspace:
	.long   0                       #  W_0
	.long   err_chan                #  Local 1
	.long   0                       #  Local 2
#endif
#ifdef HARDLINKS
maxhardchannel:  .long   MostNeg+32
	.extern in_links   /* base address of array of 8 linkdata structures */
	.extern linkdata_stride   /* size of 8 linkdata structure */
	.extern links_pending     /* number of transfers pending */
#endif
!        .globl version
!version:
!        .ascii  "KROC DEC Alpha scheduler - Version 8.5\n\0"
/*}}}  */

	.align  3                       #  For "LDS"
/*{{{  Floating-point constants in IEEE S format*/
fconst: .long   0x40000000              #  2.0
	.long   0x3f000000              #  0.5
	.long   0x4f800000              #  2^32
	.long   0x2f800000              #  2^(-32)
	.long   0x4b000000              #  2^23
	.long   0x59800000              #  2^52
/*}}}  */

	.text
/*{{{  entry points*/
! table of entry points (as jumps):
	br       $fpsqrt
	br       $ldiv
	br       $rem
	br       $div
	br       $move
	br       dummy  # $roundsn
	br       dummy  # $postnormsn
	br       dummy  # $unpacksn
	br       dummy  # $norm
	br       $seterr
	br       $rangerr
	br       $dist
	br       $disc
	br       $diss
	br       $enbt
	br       $enbc
	br       $taltwt
	br       $altwt
	br       $tin
	br       $ldtimer
	br       dummy  # $pause
	br       $stopp
	br       $endp
	br       $startp
	br       $runp
	br       $outbyte
	br       $outword
	br       $out
	br       $in
.globl  occam_table
occam_table:
dummy:
	jmp      zero,(L)
/*}}}  */
	.globl  enter
	.ent    enter
/*{{{  enter here*/
/*****************************************************************************/
/* Set up hard channels as pseudo-parameters.                                */
/* Initialize registers and enter occam program at label "$entry_user".      */
/* return error flag to calling code.                                        */
/*****************************************************************************/

enter:
	ldgp    gp,0(pv)

/*{{{  set up floating point constants*/
loadfloat:
/*****************************************************************************/
/* Set up constants in floating-point registers.                             */
/* Called in initialization.                                                 */
/* Registers F2..F9 are preserved in standard conforming procedures          */
/*****************************************************************************/

	lda     T0, fconst
	lds     F5,0(T0)               #  two
	lds     F4,4(T0)               #  half
	lds     F6,8(T0)               #  two32
	lds     F7,12(T0)              #  twom32
	lds     F8,16(T0)              #  two23
	lds     F9,20(T0)              #  two52
/*}}}  */
/*{{{  floating point magic - not done yet!*/
/* assume C runtime system has defaulted everything sensibly */
/*}}}  */

#ifndef ALONE
	ldq     T0,kroc_out
	lda     T1,scr_workspace
	stl     T0,8(T1)                #  local2 of screen process workspace
	ldq     T0,kroc_err
	lda     T1,err_workspace
	stl     T0,8(T1)                #  local2 of stderr process workspace
#endif
	ldl     MIN,mostnegval          #  Used here and in user program
	bis     MIN,MIN,Front           #  front of process queue
	bis     MIN,MIN,Back            #  back of process queue
	bis     MIN,MIN,TimeQ           #  front of timer queue
	lda     ETAB,occam_table        #  kernel entry point table
	lda     pv,occamentry
	jmp     zero,(pv),occamentry      #  Set up main parameters and run
!                                       #  user program (should not change gp)
/*}}}  */
/*{{{  kbd_process*/
#ifndef ALONE
/*****************************************************************************/
/* Process for keyboard.                                                     */
/* This occam process is scheduled when kbd_sync is found to have been set   */
/* to 1.                                                                     */
/* It then reads the signalled keystroke, sends it on and acknowledges it    */
/*****************************************************************************/

kbd_process:
#ifdef KBD_SIGNAL
 .set at
	stl     zero,kbd_sync           #  Clear kbd_sync
 .set noat
#endif
kbd_read:
	lda     pv,read_the_kbd
	jsr     ra,(pv),read_the_kbd    #

	ldgp    gp,0(ra)

#ifndef KBD_SIGNAL
 .set at
	stl     zero,kbd_sync           #  Clear kbd_sync
 .set noat
#endif
	lda     Wptr,kbd_workspace      #  kbd_workspace
	stl     R0,0(Wptr)              #  Character in workspace
	lda     L, 1f                   #  arrange to return to call ack
	lda     REG_1,kbd_chan          #  kbd_chan
	lda     REG_2,1(zero)           #  Count of one
	br      set_W                   #  Continue as $in
!       ------------------------

1:
	lda     pv,acknowledge
	jsr     ra,(pv),acknowledge     #
	ldgp    gp,0(ra)

	br      $scheduler
!       ------------------------
#endif
/*}}}  */

#ifdef HARDLINKS
/*{{{  hardoutput using C_outmessage (linkno)*/
/*****************************************************************************/
/*      REG_2 - (Transputer A) Number of bytes to transfer                   */
/*      REG_1   MININT + PLACE address of hard channel                       */
/*      REG_0 - (Transputer C) Address of source data                        */
/*****************************************************************************/
hardoutput:
	subl    REG_1,MIN,REG_1         # index transputer-size words
	bis     Wptr,Wptr,P3            # Wptr to save in channel
	srl     REG_1,3,P2              # link number
	bis     REG_0,REG_0,P1          # data pointer
	stl     REG_0,W_Pointer(Wptr)   # belt and braces
	bis     REG_2,REG_2,P0          # bytes to write
	stl     L,W_Iptr(Wptr)          # save return address
	lda     pv,C_outmessage
	jsr     ra,(pv),C_outmessage    #
	ldgp    gp,0(ra)
	beq     R0,$scheduler           # deschedule if it returns 0
	ldl     L,W_Iptr(Wptr)
	jmp     zero,(L)                # return to occam
/*}}}  */
/*{{{  hardinput using C_inmessage (linkno)*/
/*****************************************************************************/
/*      REG_2 - (Transputer A) Number of bytes to transfer                   */
/*      REG_1   MININT + PLACE address of hard channel                       */
/*      REG_0 - (Transputer C) Address of destination                        */
/*****************************************************************************/
hardinput:
	subl    REG_1,4,REG_1           # first input link is at 4
	subl    REG_1,MIN,REG_1         # index transputer-size words
	srl     REG_1,3,a2              # link number
	bis     Wptr,Wptr,P3            # Wptr to save in channel
	bis     REG_0,REG_0,P1          # data pointer
	stl     REG_0,W_Pointer(Wptr)   # belt and braces
	bis     REG_2,REG_2,P0          # bytes to read
	stl     L,W_Iptr(Wptr)          # save return address
	lda     pv,C_inmessage
	jsr     ra,(pv),C_inmessage     #
	ldgp    gp,0(ra)
	beq     R0,$scheduler           # deschedule if it returns 0
	ldl     L,W_Iptr(Wptr)
	jmp     zero,(L)                # return to occam
/*}}}  */
#endif
/*{{{  outbyte, outword and out*/
/*****************************************************************************/
/* Output from a channel.                                                    */
/*                                                                           */
/*      REG_1 (Transputer A) - Word or byte to transfer                      */
/*      REG_0 (Transputer B) - Channel address                               */
/*****************************************************************************/

	.globl $outbyte
$outbyte:
	lda     REG_2,1(zero)               #  Count of one
	br      1f
!       ------------------------

	.globl $outword
$outword:
	lda     REG_2,4(zero)               #  Count of four
1:      stl     REG_1,0(Wptr)               #  Save word or byte
	bis     REG_0,REG_0,REG_1           #  Get registers straight
set_W:
	bis     Wptr,Wptr,REG_0             #  REG_0 points to byte(s)

!       fall through

/*****************************************************************************/
/*      REG_2 - (Transputer A) Number of bytes to transfer                   */
/*      REG_1 - (Transputer B) Channel address                               */
/*      REG_0 - (Transputer C) Address of source data                        */
/*****************************************************************************/

	.globl $out
$out:
#ifdef HARDLINKS
	ldl     T0,maxhardchannel
	cmple   REG_1,T0,P5             #  Is this a hard channel?
	bne     P5,hardoutput           #  yes, go do it
#endif
	ldl     T0,0(REG_1)             #  Get other process Wptr
	cmpeq   T0,MIN,P5               #  Is it NotProcess?
	beq     P5,1f                   #  No, we can do the copy
next:
	stl     Wptr,0(REG_1)           #  Yes, Workspace into channel
	stl     REG_0,W_State(Wptr)     #  Message address into workspace
	br      $stopp

1:
	bic     T0,1,T1                 #  Get rid of priority bit to get Wptr
	ldl     T2,W_State(T1)          #  Get ALT descriptor or destination addr
	addl    MIN,Ready_d,T3          #  (Ready_p)
	cmple   T2,T3,P5                #  Special case?
	bne     P5,3f                   #  Yes
!                                       #  No - copy data

	stl     MIN,0(REG_1)            #  (NotProcess_p)
	bis     T2,T2,REG_1             #  Destination address
	br      copy                    #  Source address already in REG_0, uses T0
!       ------------------------

3:      cmpeq   T2,T3,P5
	bne     P5,next                 #  Ready

	stl     T3,W_State(T1)          #  Ready_p
	addl    MIN,Waiting_d,T3        #  (Waiting_p)
	cmpeq   T2,T3,P5
	beq     P5,next                 #  No - must be Enabling (Impossible?)

!                                       #  Yes
	stl     L,W_Iptr(Wptr)
	stl     REG_0,W_State(Wptr)
	stl     Wptr,0(REG_1)           #  Place ours in the channel
	cmpeq   Front,MIN,P5            #  Is the queue non-empty?
	beq     P5,1f                   #  Yes, link this at back

	bis     T1,T1,Front             #  No, this becomes front and back
	br      2f
1:      stl     T1,W_Link(Back)         #
2:      bis     T1,T1,Back
	br      $scheduler              #  De-schedule this process
!       ------------------------

/*}}}  */

/*{{{  output to unix stdout or stderr*/
#ifndef ALONE
/*****************************************************************************/
/* Process for hard output.                                                  */
/*****************************************************************************/

hard_print:
	/*{{{  ldbu    P0,0(Wptr)              #  Get the character from workspace - uses t9=T1*/
	ldq_u   T1,0(Wptr)
	extbl   T1,Wptr,P0
	/*}}}  */
	lda     T2,Flush(zero)
	cmpeq   P0,T2,P5                #  Is it a Flush code?
	bne     P5,1f                   #  Yes

	ldl     P1,Local2(Wptr)
	lda     pv,fputc
	jsr     ra,(pv),fputc           #  fputc (character, port)
	ldgp    gp,0(ra)
	br      2f
!       ------------------------

1:      ldl     P0,Local2(Wptr)
	lda     pv,fflush
	jsr     ra,(pv),fflush          #  fflush (port)
	ldgp    gp,0(ra)

2:
	ldl     MIN,mostnegval          #  Corrupted by fputc?
	lda     L,hard_print
	lda     REG_2,1(zero)
	ldl     REG_1,Local1(Wptr)
	bis     Wptr,Wptr,REG_0

!       fall through
#endif
/*}}}  */
/*{{{  in*/
/*****************************************************************************/
/* Input from a channel.                                                     */
/*      REG_2 - (Transputer A) Number of bytes to transfer;  > 0 (CWG p.21)  */
/*      REG_1 - (Transputer B) Channel address (hardware link address)       */
/*      REG_0 - (Transputer C) Address of destination for data               */
/*****************************************************************************/

	.globl $in
$in:
#ifdef HARDLINKS
	ldl     T0,maxhardchannel
	cmple   REG_1,T0,P5             #  Is this a hard channel?
	bne     P5, hardinput           #  yes, go do it
#endif
#ifndef ALONE
#ifndef KBD_SIGNAL
	lda     T1,kbd_chan             #  kbd_chan
	cmpeq   REG_1,T1,P5             #  is this kbd input?
	beq     P5,1f                   #  no, soft input
.set at
	stl     P5,kbd_sync             #  set kbd_sync non-zero, to force input
.set noat
1:
#endif
#endif
	ldl     T0,0(REG_1)             #  Get other process Wptr
	cmpeq   T0,MIN,P5               #  Has other party communicated?
	bne     P5,next                 #  De-schedule this process

	stl     MIN,0(REG_1)            #  Reset channel to NotProcess_p
	bis     REG_0,REG_0,REG_1       #  Destination address
	ldl     REG_0,W_State(T0)       #  Fetch source address

!       Shared by "$in" and "$out", the cost is undetectable
!       T0 points to other workspace
!                                       #  Put other process on back of queue
copy:   cmpeq   Front,MIN,P5            #  Is the queue non-empty?
	beq     P5,1f                   #  Yes, link at back
	bis     T0,T0,Front             #  No, set this as front and back
	br      2f

1:      stl     T0,W_Link(Back)         #
2:                                      #  Other process is at back of queue
	stl     Wptr,W_Link(T0)         #  Chain on current process
	stl     L,W_Iptr(Wptr)          #  Place Iptr into workspace
	bis     Wptr,Wptr,Back          #  Make it the last process
!                                       #  This process is at back of queue
	bis     Front,Front,Wptr        #  Get the first process in queue
	ldl     Front,W_Link(Front)     #  Point Front to next process
	ldl     L,W_Iptr(Wptr)          #  Make new process active

!       IN/OUT must be at least one byte (CWG p.21)
!       MOVE is unlikely to be used for a single byte

	cmpeq   REG_2,1,P5              #  Single byte?
	beq     P5,movem                #  No

	/*{{{  ldb     T0,0(REG_0)*/
	ldq_u   T1,0(REG_0)
	extbl   T1,REG_0,T0
	/*}}}  */
	/*{{{  stb     T0,0(REG_1)*/
	ldq_u   T1,0(REG_1)
	insbl   T0,REG_1,T2
	mskbl   T1,REG_1,T1
	bis     T1,T2,T1
	stq_u   T1,0(REG_1)
	/*}}}  */
	jmp     zero,(L)
!       ------------------------

/*}}}  */
/*{{{  block move*/

/*****************************************************************************/
/* Block move:                                                               */
/*      REG_2 - (Transputer A) Number of bytes to copy                       */
/*      REG_1 - (Transputer B) Destination address                           */
/*      REG_0 - (Transputer C) Source address                                */
/* Optimized for qword- and lword-aligned data (source and destination)      */
/*      If not does each end by bytecopy subroutine and middle by            */
/*      copying quadwords, or longwords or by building                       */
/*      destination quadwords from two parts each, according to alignment    */
/*      Remember that occam does not need overlapped copies!                 */
/*****************************************************************************/

/*{{{  occam version to aid comprehension*/
! -- s = REG_0
! -- d = REG_1
! -- l = REG_2
! -- d1 = P1             -- odd bytes in first destination qword
! INT T3:                -- odd bytes in final destination qword
! INT T0:                -- temporary
! SEQ
!   T0 := (s\/d)/\7      -- test for alignment of source and dest
!   IF
!     T0 = 0
!       SEQ                                     -- copyquads
!         T3 := l/\7
!         WHILE l>=8                            -- quadloop
!           SEQ
!             qcopy (s, d, 8)   -- aligned qword
!             l := l - 8
!     TRUE
!       SEQ                                     -- trywords
!         T0 := T0 /\ 3
!         IF
!           T0 = 0
!             SEQ
!               T3 := l/\3
!               WHILE l>=4                      -- wordloop
!                 SEQ
!                   lcopy (s, d, 4)  -- aligned  lword
!                   l := l - 4
!           TRUE
!             SEQ                               -- gencopy
!               IF
!                 l<8
!                   SEQ                         -- shortcopy
!                     bcopy (s, d, l) -- unaligned bytes
!                     T3 := 0         -- suppress lastfew
!                 TRUE
!                   SEQ                         -- longcopy
!                     T3 := (d+l)/\7
!                     d1 := 8-(d/\7)
!                     l := l-d1
!                     IF
!                       d1=0
!                         SKIP
!                       TRUE
!                         bcopy (s, d, d1) -- unaligned first bytes
!                     IF
!                       (s/\7) = 0
!                          WHILE l>=8          -- quadloop
!                            SEQ
!                              qcopy (s, d, 8)   -- aligned qword
!                              l := l - 8
!                       (s/\3) = 0
!                          WHILE l>=4          -- lwordloop
!                            SEQ
!                              lcopy (s, d, 4)  -- aligned  lword
!                              l := l - 4
!                       TRUE
!                         WHILE l >= 8
!                           SEQ
!                             qucopy (s, d, 8)   -- unaligned qword
!                             l := l - 8
!   IF
!     T3=0
!       SKIP
!     TRUE
!       SEQ                                     -- lastfew
!         bcopy (s, d, T3)   --unaligned last bytes

/*}}}  */
	.globl $move
$move:
movem:  bis     REG_0,REG_1,T0          #  Can we copy qwords?
	and     T0,7,T0                 #  look at bottom 3 bits of source/dest
	bne     T0,trylwords            #  if not completely quadword-aligned

copyquads:
	and     REG_2,7,T3              #  odd bytes
quadloop:
	subl    REG_2,8,REG_2           #  8 bytes each time round
	blt     REG_2,lastfew           #  copy the last T3 one-by-one
	ldq     T1,0(REG_0)             #  load 8 bytes quad aligned
	addl    REG_0,8,REG_0           #  increment source address
	stq     T1,0(REG_1)             #  store 8 bytes quad aligned
	addl    REG_1,8,REG_1           #  increment destination address
	bne     REG_2,quadloop          #  loop unless all done
	jmp     zero,(L)

trylwords:
	and     T0,3,T0                 #  bottom 2 bits, can we copy words?
	bne     T0,gencopy              #  if not word aligned
	and     REG_2,3,T3              #  odd bytes
lwordloop:
	subl    REG_2,4,REG_2           #  4 bytes each time round
	blt     REG_2,lastfew           #  copy the last T3 one-by-one
	ldl     T1,0(REG_0)             #  load 4 bytes word aligned
	addl    REG_0,4,REG_0           #  increment source address
	stl     T1,0(REG_1)             #  store 4 bytes word aligned
	addl    REG_1,4,REG_1           #  destination address
	bne     REG_2,lwordloop         #  loop unless all done
	jmp     zero,(L)

gencopy:
	cmple   REG_2,8,P5              #  are we copying <= 8 bytes?
	beq     P5,longcopy             #  No, do most of it by ldqu/stq within string

shortcopy:
	bis     L,L,ra
	br      byteloop                #  do it a byte at a time

longcopy:
	addl    REG_1,REG_2,T3          #  address of 1st byte not to be overwritten
	and     T3,7,T3                 #  odd bytes in final destination qword
	bis     zero,8,P4
	and     REG_1,7,T0              #  unused bytes in 1st destination qword
	subl    P4,T0,P1                #  destination bytes upto 1st qword boundary
	subl    REG_2,P1,P2             #  other bytes to be copied (qword aligned dest)
	beq     P1,1f
	bis     P1,P1,REG_2             #  for the subroutine
	bsr     ra,byteloop             #  copy leading bytes
1:
	bis     P2,P2,REG_2             #  recover the count
	and     REG_0,7,P5              #  is source quad aligned?
	bne     P5,3f                   #  No, try lwords
	br      quadloop                #  Yes, aligned are quicker
3:      and     REG_0,3,P5              #  is source lword aligned
	bne     P5,2f                   #  No, use unaligned loads
	br      lwordloop               #  Yes, lwords better than bytes

2:      subl    REG_2,8,REG_2           #  copy 8 bytes each time round
	blt     REG_2,lastfew           #  less than 8 => go to last few special case
/*{{{  load unaligned quadword from address REG_0*/
	ldq_u   T0,0(REG_0)
	ldq_u   T1,7(REG_0)
	extql   T0,REG_0,T0
	extqh   T1,REG_0,T1
	bis     T0,T1,T0
/*}}}  */
	stq     T0,0(REG_1)              #  to destination (aligned qword)
	addl    REG_0,8,REG_0            #  increment source address
	addl    REG_1,8,REG_1            #  increment destination address
	br      2b

lastfew:
	bis     L,L,ra                   #  arrange to return direct to occam
	beq     T3,return                #  destination ended at qword boundary
	bis     T3,T3,REG_2              #  odd bytes in final destination qword

byteloop:
! copies REG_2 bytes from (REG_0) et seq to (REG_1) et seq
! on exit REG_2 has decremented to zero, REG_0 and REG_1 have incremented by REG_2
! note that ldb uses t9 = T1, stb uses t9,t10 = T1,T2
	/*{{{  ldb     T0,0(REG_0)*/
	ldq_u   T1,0(REG_0)
	extbl   T1,REG_0,T0
	/*}}}  */
	addl    REG_0,1,REG_0
	/*{{{  stb     T0,0(REG_1)*/
	ldq_u   T1,0(REG_1)
	insbl   T0,REG_1,T2
	mskbl   T1,REG_1,T1
	bis     T1,T2,T1
	stq_u   T1,0(REG_1)
	/*}}}  */
	addl    REG_1,1,REG_1
	subl    REG_2,1,REG_2
	bgt     REG_2,byteloop
return:
	jmp     zero,(ra)
!       ------------------------
/*}}}  */

/*{{{  stopp and scheduler*/
/*****************************************************************************/
/* This function de-schedules the current process. If the process   */
/* queue contains one or more runnable processes, then run the next process; */
/* otherwise wait for external event (kbd, time, link) or finish.                                                         */
/*****************************************************************************/

	.globl $resched                 #  should only come here if a sync is set
$resched:
	cmpeq   Front,MIN,P5            #  Is the queue non-empty?
	beq     P5,1f                   #  Yes, link this at back

	bis     Wptr,Wptr,Front         #  No, this becomes front and back
	br      2f
1:      stl     Wptr,W_Link(Back)       #  Fall through
2:      bis     Wptr,Wptr,Back

	.globl $stopp
$stopp:                                 #  STOPP, ALTWT, STOPERR
	stl      L,W_Iptr(Wptr)         #  Save Iptr in workspace

!       fall through

	.globl $scheduler
$scheduler:                             #  ENDP and internal
/*{{{  if there are hard links check them*/
#ifdef HARDLINKS
	ldl     T0, link_sync           #  Any link transfers finished?
	beq     T0, links_none
/*{{{  reschedule any processes whose link transfers have completed*/
!-----------------------------------------------------------
!   SEQ i = 0 FOR ins
!     IF
!       in_links[i][flag]
!         ... put in_links[i][wptr] on queue
!       TRUE
!         SKIP
!
!   SEQ i = 0 FOR outs
!     IF
!       out_links[i][flag]
!         ... put out_links[i][wptr] on queue
!       TRUE
!         SKIP
!-----------------------------------------------------------
!  This simple version assumes there are 8 structures
!  stored adjacently starting at in_links
!  Each structure has size linkdata_stride and includes fields:
!  flag at byte 4, wptr (occam channel) at byte 8
!
	bis     zero,8,T3
  .set at
	ldl     T0,in_links                    #  address of extlinks[0]
  .set noat
lloop:
	ldl     T1,4(T0)                       #  flag word
	beq     T1,1f                          #  do nothing if zero
	stl     zero,4(T0)                     #  clear flag if set
	ldl     T1,8(T0)                       #  wptr word
	cmpeq   Front,MIN,P5                   #  Is the queue empty?
	bne     P5,2f                          #  Yes, set front pointer
	stl     T1,W_Link(Back)                #  No, link this at back
	br      3f
2:
	bis     T1,T1,Front
3:
	bis     T1,T1,Back
  .set at
	ldl     T1,links_pending               #  decrement count of link
	subl    T1,1,T1                        #  transfers pending
	stl     T1,links_pending
  .set noat
	beq     T1,links_done                  #  short cut!
1:
  .set at
	ldl     T2,linkdata_stride
  .set noat
	addl    T0,T2,T0                       #  add length of struct
	subl    T3,1,T3                        #  count down
	bne     T3,lloop
links_done:
  .set at
	stl     zero,link_sync          #  Mark link rescheduling done
  .set noat
/*}}}  */
links_none:
#endif
/*}}}  */
#ifndef ALONE
	ldl     T0,kbd_sync             #  Get kbd_sync
	cmpeq   T0,1,P5                 #  is it set?
#ifdef KBD_SIGNAL
	bne     P5,kbd_process
#else
	beq     P5,1f                   #  kbd_sync set if a key awaited
	lda     pv, test_the_kbd
	jsr     ra,(pv),test_the_kbd    #
	ldgp    gp,0(ra)
	beq     R0,s.time               # no key, look at clock
	br      kbd_read                # in keyboard process code
1:
#endif
#endif
#ifdef TIM_SIGNAL
	ldl     T1,tim_sync             #  Get tim_sync
	cmpeq   T1,1,P5                 #  Timer gone off?
	beq     P5,s.queue              #  No
  .set at
	stl     zero,tim_sync           #  Clear tim_sync
  .set noat
#endif
/*{{{  moving processes from timer to run queue*/
/*****************************************************************************/
/*  SEQ                                                                      */
/*    tim ? now                                                              */
/*    removed := FALSE                                                       */
/*    WHILE (TimeQ <> NotProcess.p) and (NOT (TimeQ^[W_Time] AFTER now))     */
/*      SEQ                                                                  */
/*        ...  move first process on timer queue to the run queue            */
/*        removed := TRUE                                                    */
/*    IF                                                                     */
/*      (TimeQ <> NotProcess.p) and removed                                  */
/*        ualarm (TimeQ^[W_Time] MINUS now, 0)                               */
/*      TRUE                                                                 */
/*        SKIP                                                               */
/*****************************************************************************/
s.time:

/*{{{  GETTIME*/
#if defined(HOST_OS_OSF30)
	lda      pv,rtime
	jsr      ra,(pv),rtime
	ldgp     gp,0(ra)
#else
	rpcc     R0
	zapnot   R0,0xf,R0
#endif
/*}}}  */

	bis     zero,zero,T0            #  removed := FALSE
s.loop:
	cmpeq   TimeQ,MIN,P5            #  WHILE TimeQ <> NotProcess_p ...
	bne     P5,s.queue              #  Short cut

	ldl     T1,W_Time(TimeQ)        #  ... and NOT (TimeQ^[W_Time] AFTER now)

	subl    T1,R0,P5                #  then MINUS now
	bgt     P5,done                 #  exit loop if then AFTER now

/*{{{  Move first process on timer queue to the run queue*/
/*****************************************************************************/
/*                                                                           */
/*  POINTER temp := TimeQ:                                                   */
/*  SEQ                                                                      */
/*    TimeQ := TimeQ^[W_TLink]                                               */
/*    temp^[W_TLink] := TimeSet.p         -- i.e. time expired               */
/*    temp^[W_Time] := now                -- ??                              */
/*    IF                                                                     */
/*      temp^[W_State] = Ready.p                                             */
/*        SKIP                            -- (was impossible (2))            */
/*      TRUE                                                                 */
/*        SEQ                                                                */
/*          temp^[W_State] := Ready.p                                        */
/*          ...  add on to run queue                                         */
/*****************************************************************************/
	bis     TimeQ,TimeQ,T1                 #  temp := TimeQ
	ldl     TimeQ,W_TLink(TimeQ)           #  Remove from timer queue
	addl    MIN,TimeSet_d,T2               #  temp^[W_TLink] := TimeSet.p
	stl     T2,W_TLink(T1)
	stl     R0,W_Time(T1)                  #  temp^[W_Time] := now
	addl    MIN,Ready_d,T2                 #  temp^[W_State] = Ready.p?
	ldl     T3,W_State(T1)
	cmpeq   T2,T3,P5
	bne     P5,2f
1:
	stl     T2,W_State(T1)                 #  temp^[W_State] := Ready.p
	cmpeq   Front,MIN,P5                   #  Is the queue empty?
	bne     P5,setfront                    #  Yes, set front pointer
	stl     T1,W_Link(Back)                #  No, link this at back
	br      setback
setfront:
	bis     T1,T1,Front
setback:
	bis     T1,T1,Back
2:
/*}}}  */

	lda     T0,True(zero)                  # removed := TRUE
	br      s.loop
!       ------------------------

done:
	lda     P0,6(zero)
!       beq     T0,badalarm             #  IF not removed

#ifdef HOST_OS_OSF30
	subl    T1,R0,P0                #  Set alarm
	bis     zero,zero,P1
	lda     pv,ualarm
	jsr     ra,(pv),ualarm
	ldgp    gp,0(ra)
#else
	bis     T1,T1,P0                #  alarm time
	bis     R0,R0,P2                #  time now
	lda     pv,setalarm
	jsr     ra,(pv),setalarm
	ldgp    gp,0(ra)
#endif
/*}}}  */
/*{{{  schedule the front process from the queue*/
s.queue:
	cmpeq   Front,MIN,P5            #  Any processes in queue?
	beq     P5,2f                   #  Yes

#ifndef ALONE
	ldl     T0,kbd_chan             #  kbd_chan
	cmpeq   T0,MIN,P5               #  Empty?
	beq     P5,1f                   #  No - wait for key
#endif

	cmpeq   TimeQ,MIN,P5            #  Yes - Timer queue empty?
	beq     P5,1f                   #  Yes - No more processes to run
#ifdef HARDLINKS
  .set at
	ldl     T1,links_pending        #  count of link transfers pending
  .set noat
	bne     T1,1f                   #  URGENT thread will set link_sync
#endif
	bis     zero,zero,P0
	lda     pv,dead
	jsr     ra,(pv),dead
!       ------------------------

1:      lda     pv,safe_pause           #  No - wait for key/timeout (signal)
	jsr     ra,(pv),safe_pause
	ldgp    gp,0(ra)
	br      $scheduler
!       ------------------------

2:      bis     Front,Front,Wptr        #  The queue is not empty, take front
	ldl     L,W_Iptr(Wptr)
	cmpeq   Front,Back,P5           #  Only one element?
	beq     P5,1f                   #  No
	bis     MIN,MIN,Back            #  Make Back null
	bis     MIN,MIN,Front           #  Make Front null
	jmp     zero,(L)
!       ------------------------

1:      ldl     Front,W_Link(Front)     #  make next Front
	jmp     zero,(L)                #  The workspace pointer is set up.
!                                       #  Just need to pass control to the
!                                       #  new current process and move
!                                       #  the front pointer on.
!       ------------------------
/*}}}  */
/*}}}  */
/*{{{  tin*/

/*****************************************************************************/
/* Timeout.                                                                  */
/*                                                                           */
/*      REG_0 (Transputer A) - Time.                                         */
/*                                                                           */
/*  Only real (not user) time.                                               */
/*****************************************************************************/

	.globl $tin
$tin:
/*{{{  GETTIME*/
#if defined(HOST_OS_OSF30)
	stl      REG_0,W_Time(Wptr)     #  rtime may corrupt it
	lda      pv,rtime
	jsr      ra,(pv),rtime
	ldgp     gp,0(ra)
	ldl      REG_0,W_Time(Wptr)
#else
	rpcc     R0
	zapnot   R0,0xf,R0
#endif
/*}}}  */

	subl    R0,REG_0,P5             #  now AFTER alarmtime
	bgt     P5,t.end                #  Yes - return

!                                       #  No - wait:
	stl     L,W_Iptr(Wptr)          #  save link back to occam
	addl    MIN,Waiting_d,T0        #  T9000 book says "Enabling"
	stl     T0,W_State(Wptr)
	addl    REG_0,1,REG_0           #  T9000 book again
	stl     REG_0,W_Time(Wptr)

/*{{{  add to timer queue*/
/*****************************************************************************/
/*  POINTER current := TimeQ:                                                */
/*  POINTER previous := NotProcess.p:                                        */
/*  SEQ                                                                      */
/*    WHILE (current <> NotProcess.p) and (timeout AFTER current^[W_Time])   */
/*      SEQ                                                                  */
/*        previous := current           -- deduce: previous <> NotProcess.p  */
/*        current := current^[W_TLink]                                       */
/*    Wptr^[W_TLink] := current                                              */
/*    IF                                                                     */
/*      previous <> NotProcess.p        -- iff: current <> TimeQ             */
/*        SEQ                           -- insert in middle of queue         */
/*          previous^[W_TLink] := Wptr                                       */
/*      TRUE                                                                 */
/*        SEQ                           -- insert at head of queue           */
/*          TimeQ := Wptr                                                    */
/*          --tim ? now                 -- we already have it                */
/*          ualarm (timeout MINUS now, 0)                                    */
/*****************************************************************************/

	bis     TimeQ,TimeQ,T0          #  current := TimeQ
	bis     MIN,MIN,T1              #  previous := NotProcess.p
tqloop: cmpeq   T0,MIN,P5               #  WHILE (current <> MIN) ...
	bne     P5,1f

	ldl     T2,W_Time(T0)           #  ... and (REG_0 AFTER current^[W_Time])
	subl    REG_0,T2,P5             #  timeout MINUS queuetime
	ble     P5,1f                   #  exit loop if not AFTER

	bis     T0,T0,T1                #  previous := current
	ldl     T0,W_TLink(T0)          #  current := current^[W_TLink]
	br      tqloop
!       ------------------------

1:      stl     T0,W_TLink(Wptr)        #  Wptr^[W_TLink] := current
	cmpeq   T1,MIN,P5               #  IF previous = MIN, do not link
	bne     P5,2f
	stl     Wptr,W_TLink(T1)        #  previous^[W_TLink] := Wptr
	br      $scheduler

2:      bis     Wptr,Wptr,TimeQ         #  ELSE TimeQ := Wptr
#ifdef HOST_OS_OSF30
	subl    REG_0,R0,P0             #  Set alarm
	bis     zero,zero,P1            #  clear P1
	lda     pv,ualarm
	jsr     ra,(pv),ualarm
	ldgp    gp,0(ra)
#else
	bis     REG_0,REG_0,P0          #  alarm time
	bis     R0,R0,P2                #  time now
	lda     pv,setalarm
	jsr     ra,(pv),setalarm
	ldgp    gp,0(ra)
#endif
	br      $scheduler
!       ------------------------

t.end:
	jmp     zero,(L)                      #  return to occam
!       ------------------------
/*}}}  */
/*}}}  */
/*{{{  taltwt*/
/*****************************************************************************/
/*  Timer ALT Wait.                                                          */
/*****************************************************************************/

	.globl $taltwt
$taltwt:

/*****************************************************************************/
/*  see T9000 transputer instruction set manual p436                         */
/*  SEQ                                                                      */
/*    Wptr^[W_0] := NoneSelected.o                                           */
/*    IF                                                                     */
/*      Wptr^[W_State] = Ready.p                                             */
/*        SEQ                                   -- why?                      */
/*          tim ? now                                                        */
/*          Wptr^[W_Time] := now                                             */
/*      Wptr^[W_TLink] = TimeNotSet.p                                        */
/*        SEQ                                                                */
/*          Wptr^[W_State] = Waiting.p            -- see altwt               */
/*          ...  wait until Wptr^[W_State] = Ready.p                         */
/*      now AFTER Wptr^[W_Time]              -- corrected MDP 5/12/96        */
/*        SEQ                                                                */
/*          Wptr^[W_State] := Ready.p                                        */
/*          Wptr^[W_Time] := now                  -- why not?                */
/*      TRUE                                                                 */
/*        SEQ                                                                */
/*          Wptr^[W_Time] := Wptr^[W_Time] + 1                               */
/*          ...  insert this process in timer queue at incremented time      */
/*          IF                                                               */
/*            Wptr^[W_State] <> Ready.p           -- redundant?              */
/*              SEQ                                                          */
/*                Wptr^[W_State] := Waiting.p     -- see altwt               */
/*                ...  wait until Wptr^[W_State] = Ready.p                   */
/*            TRUE                                                           */
/*              impossible (1)                                               */
/*****************************************************************************/

	lda     T0,NoneSelected_o(zero)    #  Wptr^[W_0] := NoneSelected.o
	stl     T0,W_0(Wptr)
	addl    MIN,Ready_d,T0
	ldl     T1,W_State(Wptr)
	cmpeq   T0,T1,P5
	beq     P5,1f                      #  state <> Ready

/*{{{  GETTIME*/
#if defined(HOST_OS_OSF30)
	lda      pv,rtime
	jsr      ra,(pv),rtime
	ldgp     gp,0(ra)
#else
	rpcc     R0
	zapnot   R0,0xf,R0
#endif
/*}}}  */
	stl     R0,W_Time(Wptr)            #  Wptr^[W_Time] := now
	jmp     zero,(L)
!       ------------------------

1:      addl    MIN,TimeNotSet_d,T0
	ldl     T1,W_TLink(Wptr)
	cmpeq   T0,T1,P5
	beq     P5,1f                      #  Wptr^[W_TLink] <> TimeNotSet.p

	addl    MIN,Waiting_d,T0
	stl     T0,W_State(Wptr)
	br      $stopp                     #  deschedule this process
!       ------------------------

1:
/*{{{  GETTIME*/
#if defined(HOST_OS_OSF30)
	lda      pv,rtime
	jsr      ra,(pv),rtime
	ldgp     gp,0(ra)
#else
	rpcc     R0
	zapnot   R0,0xf,R0
#endif
/*}}}  */
	ldl     REG_0,W_Time(Wptr)         #  use REG_0 as temp for timeout
	subl    REG_0,R0,P5                #  timeout MINUS now
	bge     P5,1f                      #  not there yet, do not set ready

	addl    MIN,Ready_d,T0             #
	stl     T0,W_State(Wptr)
	stl     R0,W_Time(Wptr)
	jmp     zero,(L)
!       ------------------------

1:      addl    REG_0,1,REG_0
	stl     REG_0,W_Time(Wptr)
/*{{{  insert on timer queue*/
/*****************************************************************************/
/*  -- Insert process in timer queue.                                        */
/*                                                                           */
/*  POINTER current := TimeQ:                                                */
/*  POINTER previous := NotProcess.p:                                        */
/*  SEQ                                                                      */
/*    WHILE (current <> NotProcess.p) and (timeout AFTER current^[W_Time])   */
/*      SEQ                                                                  */
/*        previous := current           -- deduce: previous <> NotProcess.p  */
/*        current := current^[W_TLink]                                       */
/*    Wptr^[W_TLink] := current                                              */
/*    IF                                                                     */
/*      previous <> NotProcess.p        -- iff: current <> TimeQ             */
/*        SEQ                           -- insert in middle of queue         */
/*          previous^[W_TLink] := Wptr                                       */
/*      TRUE                                                                 */
/*        SEQ                           -- insert at head of queue           */
/*          TimeQ := Wptr                                                    */
/*          ualarm (timeout MINUS now, 0)                                    */
/*****************************************************************************/

	bis     TimeQ,TimeQ,T0          #  current := TimeQ
	bis     MIN,MIN,T1              #  previous := NotProcess.p
ta.loop:
	cmpeq   T0,MIN,P5               #  WHILE (current <> MIN) ...
	bne     P5,1f

	ldl     T2,W_Time(T0)           #  ... and (REG_0 AFTER current^[W_Time])
	subl    REG_0,T2,P5             #  time MINUS queuetime
	blt     P5,1f                   #  exit loop if not AFTER

	bis     T0,T0,T1                #  previous := current
	ldl     T0,W_TLink(T0)          #  current := current^[W_TLink]
	br      ta.loop
!       ------------------------

1:      stl     T0,W_TLink(Wptr)        #  Wptr^[W_TLink] := current
	cmpeq   T1,MIN,P5               #  IF previous <> MIN
	bne     P5,2f
	stl     Wptr,W_TLink(T1)        #  previous^[W_TLink] := Wptr
	br      3f

2:      bis     Wptr,Wptr,TimeQ         #  ELSE TimeQ := Wptr

#ifdef HOST_OS_OSF30
	subl    REG_0,R0,P0             #  Set alarm
	bis     zero,zero,P1            #  clear P1
	lda     pv,ualarm
	jsr     ra,(pv),ualarm
	ldgp    gp,0(ra)
#else
	bis     REG_0,REG_0,P0          #  alarm time
	bis     R0,R0,P2                #  time now
	lda     pv,setalarm
	jsr     ra,(pv),setalarm
	ldgp    gp,0(ra)
#endif

3:      addl    MIN,Ready_d,T0
	ldl     T1,W_State(Wptr)
	cmpeq   T0,T1,P5
	beq     P5,1f
	bis     L,L,P1                  #  return address to occam
	lda     P0,2(zero)
	br      imposs
1:
	addl    MIN,Waiting_d,T0        #  Wptr^[W_State] = {T}Waiting.p
	stl     T0,W_State(Wptr)
	br      $stopp
!       ------------------------

/*}}}  */
/*}}}  */
/*{{{  dist*/
/*****************************************************************************/
/*  Disable Timer.                                                           */
/*                                                                           */
/*     On input:                               On output:                    */
/*     REG_2 (Transputer A) code.offset                                      */
/*     REG_1 (Transputer B) pre.condition                                    */
/*     REG_0 (Transputer C) timeout            REG_0 (Transputer A) - result */
/*****************************************************************************/

	.globl $dist
$dist:

/*****************************************************************************/
/*  IF                                                                       */
/*    NOT pre.condition                                                      */
/*      result := FALSE                                                      */
/*    Wptr^[W_TLink] = TimeNotSet.p                                          */
/*      result := FALSE                                                      */
/*    Wptr^[W_TLink] = TimeSet.p                                             */
/*      -- some timer guard has gone off                                     */
/*      IF                                                                   */
/*        NOT (Wptr^[W_Time] AFTER timeout                                   */
/*          result := FALSE             -- it wasn't this guard              */
/*        Wptr^[W_0] <> NoneSelected.o                                       */
/*          result := FALSE             -- an earlier one was selected       */
/*        TRUE                                                               */
/*          SEQ                                                              */
/*            Wptr^[W_0] := code.offset -- this is the first ready guard     */
/*            result := TRUE                                                 */
/*    TRUE                           -- no timer guard ready; cancel timeout */
/*      result := FALSE                                                      */
/*      ... remove process from timer queue                                  */
/*****************************************************************************/

	beq     REG_1,1f                #  Pre-condition FALSE?

	addl    MIN,TimeNotSet_d,T0     #  Timeout cancelled?
	ldl     T1,W_TLink(Wptr)
	cmpeq   T0,T1,P5
	bne     P5,1f                   #  Yes

	addl    MIN,TimeSet_d,T0        #  Has some timer guard gone off?
	cmpeq   T0,T1,P5
	beq     P5,d.queue              #  No

	ldl     T0,W_Time(Wptr)         #  Was it our guard?
	subl    T0,REG_0,P5             #  queuetime MINUS timeout
	ble     P5,1f                   #  Not AFTER timeout

	ldl     T0,W_0(Wptr)            #  Was an earlier guard selected?
	lda     T3,NoneSelected_o(zero)
	cmpeq   T0,T3,P5
	beq     P5,1f                   #  Yes

!       It was this guard, and this is the first ready one

	stl     REG_2,W_0(Wptr)         #  Wptr^[W_0] := code.offset
	lda     REG_0,True(zero)
	jmp     zero,(L)
!       ------------------------

1:      bis     zero,zero,REG_0
	jmp     zero,(L)                       #  return FALSE
!       ------------------------

/*{{{  remove process from timer queue*/
/*****************************************************************************/
/*  -- No timer guard ready; cancel timeout                                  */
/*  -- Remove process from timer queue.                                      */
/*                                                                           */
/*  SEQ                                                                      */
/*    IF                                                                     */
/*      TimeQ = Wptr                     -- is it head of timer queue?       */
/*        SEQ                                                                */
/*          TimeQ := TimeQ^[W_TLink]        -- remove head of timer queue    */
/*          IF                                                               */
/*            TimeQ <> NotProcess.p      -- is timer queue not empty?        */
/*              SEQ                                                          */
/*                -- TimeQ^[W_Link] := NotProcess.p                          */
/*                IF                                                         */
/*                  (TimeQ^[W_Time] MINUS now) <= 0                          */
/*                    tim_sync := 1     -- time-out has happened#            */
/*                  TRUE                                                     */
/*                    ualarm (TimeQ^[W_Time] MINUS now, 0)                   */
/*            TRUE                                                           */
/*              SKIP                                                         */
/*      TimeQ = NotProcess.p           -- implies process not on timer queue */
/*        impossible (3)                                                     */
/*      TRUE                                                                 */
/*        -- POINTER previous := Wptr^[W_Link]:                              */
/*        -- next := Wptr^[W_TLink]                                          */
/*        -- IF                                                              */
/*        --   next <> NotProcess.p                                          */
/*        --     next^[W_Link] := previous                                   */
/*        --   TRUE                                                          */
/*        --     SKIP                                                        */
/*        -- previous^[W_TLink] := next                                      */
/*        POINTER previous := TimeQ:  -- search from 2nd item on timer queue */
/*        POINTER current := TimeQ^[W_TLink]:                                */
/*        SEQ                                                                */
/*          WHILE (current <> NotProcess.p) and (current <> Wptr)            */
/*            SEQ                                                            */
/*              previous := current                                          */
/*              current := current^[W_TLink]                                 */
/*          IF                                                               */
/*            current = NotProcess.p                                         */
/*              impossible (4)                                               */
/*            TRUE                                                           */
/*              SKIP                                                         */
/*          -- deduce: current = Wptr                                        */
/*          next := Wptr^[W_TLink]                                           */
/*          previous^[W_TLink] := next                                       */
/*    Wptr^[W_TLink] := Timed.end.p      -- i.e. timer cancelled             */
/*****************************************************************************/

d.queue:
	cmpeq   TimeQ,Wptr,P5            #  Head of timer queue?
	beq     P5,2f                    #  No

	ldl     TimeQ,W_TLink(TimeQ)     #  Remove head of timer queue
	cmpeq   TimeQ,MIN,P5             #  Timer queue empty?
	bne     P5,d.end                 #  Yes

/*{{{  GETTIME*/
#if defined(HOST_OS_OSF30)
	lda      pv,rtime
	jsr      ra,(pv),rtime
	ldgp     gp,0(ra)
#else
	rpcc     R0
	zapnot   R0,0xf,R0
#endif
/*}}}  */
	ldl     T1,W_Time(TimeQ)       # queuetime

	subl    R0,T1,P5               #  queuetime MINUS now
	bgt     P5,1f                  #  not there yet, go set alarm

	lda     T1,1(zero)
  .set at
	stl     T1,tim_sync
  .set noat
	br      d.end
!       ------------------------

#ifdef HOST_OS_OSF30
1:      subl    T1,R0,P0                 #
	bis     zero,zero,P1             #
	lda     pv,ualarm
	jsr     ra,(pv),ualarm
	ldgp    gp,0(ra)
#else
1:      bis     T1,T1,P0                 # alarm time
	bis     R0,R0,P2                 # time now
	lda     pv,setalarm
	jsr     ra,(pv),setalarm
	ldgp    gp,0(ra)
#endif
	br      d.end
!       ------------------------

2:      bis     L,L,P1                  #  return address to occam
	cmpeq   TimeQ,MIN,P5            #  Process not there?
	beq     P5,1f
	lda     P0,3(zero)
	br      imposs
1:
!       Search for Wptr after head of queue

	bis     TimeQ,TimeQ,T0          #  previous := TimeQ
	ldl     T1,W_TLink(TimeQ)       #  current := TimeQ^[W_TLink]
d.loop:
	cmpeq   T1,MIN,P5               #  WHILE current <> NotProcess.p ...
	beq     P5,3f
	lda     P0,4(zero)
	br      imposs
3:
	cmpeq   T1,Wptr,P5              #  ... and current <> Wptr
	bne     P5,1f

	bis     T1,T1,T0                #  previous := current
	ldl     T1,W_TLink(T1)          # current := current^[W_TLink]
	br      d.loop
!       ------------------------

!       We have found it, so remove it from the queue

1:      ldl     T2,W_TLink(Wptr)        #  next := Wptr^[W_TLink]
	stl     T2,W_TLink(T0)          #  previous^[W_TLink] := next

d.end:
	addl    MIN,TimeNotSet_d,T0     #  Wptr^[W_TLink] := TimeNotSet.p
	stl     T0,W_TLink(Wptr)        #
	bis     zero,zero,REG_0         #  return FALSE
	jmp     zero,(L)
!       ------------------------
/*}}}  */
/*}}}  */
/*{{{  fp sqrt*/
!  on entry:  FA  is value.  
!  on exit:   FA  is result.
	.globl  $fpsqrt
$fpsqrt:
!  may be called inside an expression, so REG0..2 must be preserved
	lda     sp,-24(sp)
	stq     REG_0,0(sp)
	stq     REG_1,8(sp)
	stq     REG_2,16(sp)
	cpys    FA,FA,F16
	lda     pv,sqrt
	jsr     ra,(pv),sqrt
	ldgp    gp,0(ra)
	cpys    F0,F0,FA
	ldq     REG_2,16(sp)
	ldq     REG_1,8(sp)
	ldq     REG_0,0(sp)
	lda     sp,24(sp)
	jmp     zero,(L)
/*}}}  */
/*{{{  moresched -- additional code which was done inline in Sparc-KROC*/
! This file contains additional code that may be called as subroutines from
! code compiled for the DEC Alpha from occam.
! for insertion into sched.S

!    Written by Michael Poole 23/March/1995
!
!-------------------------------------------------------------------------
/*{{{  ldiv*/
!  on entry: REG_2 (Transputer A) is divisor
!            REG_1 (Transputer B) is dividend (lo)
!            REG_0 (Transputer C) is dividend (hi)
!  on exit:  REG_1 (Transputer A) is quotient
!            REG_0 (Transputer B) is remainder

	.globl  $ldiv
$ldiv:
	zapnot  REG_1, 0xf, REG_1   /* unsigned */
	zapnot  REG_2, 0xf, T3      /* t11 */
	sll     REG_0, 32, T2       /* t10 */
	bis     REG_1, T2, T2
	lda     pv,__divqu
	jsr     T1,(pv),__divqu     /* assume does not corrupt REG_0,1,2 */
	ldgp    gp,0(T1)
	zapnot  REG_2, 0xf, T3      /* t11 */
	sll     REG_0, 32, T2       /* t10 */
	bis     REG_1, T2, T2
	bis     t12,t12,REG_1       /* quotient */
	lda     pv,__remqu
	jsr     T1,(pv),__remqu     /* assume does not corrupt REG_1 */
	ldgp    gp,0(T1)
	bis     t12,t12,REG_0       /* remainder */
	jmp     zero,(L)
/*}}}  */
/*{{{  div*/
! on entry: t10 (T2) is dividend
!           t11 (T3) is divisor
! on exit:  t12 is quotient
	.globl  $div
$div:
	lda     pv,__divl
	jsr     T1,(pv),__divl      /* assume does not corrupt t3 */
	ldgp    gp,0(T1)
	jmp     zero,(L)
/*}}}  */
/*{{{  rem*/
! on entry: t10 (T2) is dividend
!           t11 (T3) is divisor
! on exit:  t12 is remainder
	.globl  $rem
$rem:
	lda     pv,__reml
	jsr     T1,(pv),__reml      /* assume does not corrupt t3 */
	ldgp    gp,0(T1)
	jmp     zero,(L)
/*}}}  */
/*{{{  enbc*/
!  on entry: t11 (T3) is boolean guard
!            t10 (T2) is channel address
!  on exit:  t12 is boolean guard
	.globl  $enbc
$enbc:
	beq     T3,continue
	ldl     T0,0(T2)                # channel word
	cmpeq   MIN,T0,P5               # NotProcess ?
	beq     P5,2f                   # No, continue
	stl     Wptr,0(T2)              # Yes, save Wptr in channel
	br      continue
2:      cmpeq   T0,Wptr,P5              # not the first enable on this channel?
	bne     P5,continue
	addl    MIN,Ready_d,T1
	stl     T1,W_State(Wptr)        # State := Ready
continue:
	bis     T3,T3,t12               # move boolean guard
	jmp     zero,(L)
! -------------------------------------------------------------------------

/*}}}  */
/*{{{  enbt*/
!  on entry: t11 (T3) is boolean guard
!            t10 (T2) is wait time
!  on exit:  t12 is boolean guard
	.globl  $enbt
$enbt:
	beq     T3,enbtexit
	ldl     T0,W_TLink(Wptr)
	addl    MIN,TimeNotSet_d,T1
	cmpeq   T0,T1,P5                # W[W_TLink] = Timenotset?
	addl    MIN,TimeSet_d,T1
	beq     P5,1f
	stl     T1,W_TLink(Wptr)        # W[W_TLink] := Timeset
	br      settime
1:
	cmpeq   T0,T1,P5                # W[W_TLink] = Timeset?
	beq     P5,enbtexit
	ldl     T0,W_Time(Wptr)         # queuetime
	subl    T2,T0,P5                # wait time MINUS queuetime
	bgt     P5,enbtexit             # do nothing if wait time is AFTER
settime:
	stl     T2,W_Time(Wptr)         # put the new (earlier) time in queue
enbtexit:
	bis     T3,T3,t12
	jmp     zero,(L)
! -------------------------------------------------------------------------
/*}}}  */
/*{{{  diss*/
!  on entry: t11 (T3) is process address
!            t10 (T2) is boolean guard
!  on exit:  t12 is "selected" flag
	.globl  $diss

$diss:
	beq     T2,returnfalse
	ldl     T0,0(Wptr)
	lda     T1,NoneSelected_o(zero)
	cmpeq   T0,T1,P5                    # W[0] = empty?
	bne     P5,1f
returnfalse:
	bis     zero,zero,t12                # selected-flag := false
	jmp     zero,(L)
1:      stl     T3,0(Wptr)                   # W[0] := codeptr
	lda     t12,1(zero)                  # selected-flag := true
	jmp     zero,(L)
! -------------------------------------------------------------------------
/*}}}  */
/*{{{  startp*/
! on entry: REG_1 (Transputer A) is other process workspace,
!           REG_0 (Transputer B) is relative iptr in that process
	.globl  $startp
$startp:
	addl    L,REG_0,T1              #  Absolute Iptr of other process
	bic     REG_1,3,REG_0           #  mask off bottom 2 bits
	stl     T1,W_Iptr(REG_0)
! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
/*}}}  */
/*{{{  runp*/
! on entry: REG_0 (Transputer A) is other process workspace,
	.globl  $runp
$runp:
	bic     REG_0,3,REG_0           #  mask off bottom 2 bits
	cmpeq   Front,MIN,P5            #  Is the queue non-empty?
	beq     P5,1f                   #  Yes, link this at back

	bis     REG_0,REG_0,Front       #  No, this becomes front and back
	br      2f
1:      stl     REG_0,W_Link(Back)
2:      bis     REG_0,REG_0,Back
	jmp     zero,(L)
!-------------------------------------------------------------------------

/*}}}  */
/*{{{  endp*/
! on entry: REG_0 (Transputer A) points to successor workspace
!                  in W[1] of which there is a process count
	.globl  $endp
$endp:
	bis     REG_0,REG_0,Wptr        # successor Wptr
	ldl     T1,Local1(Wptr)
	subl    T1,1,T1
	stl     T1,Local1(Wptr)         # decrement W[1]
	bne     T1,$scheduler           # some left, go schedule
	ldl     L,0(Wptr)
	jmp     zero,(L)                # go to successor
! -------------------------------------------------------------------------
/*}}}  */
/*{{{  ldtimer*/
!  on entry: no parameters
!  on exit REG_0 (Transputer A) has current time
	.globl  $ldtimer
$ldtimer:
/*{{{  GETTIME*/
#if defined(HOST_OS_OSF30)
	lda      pv,rtime
	jsr      ra,(pv),rtime
	ldgp     gp,0(ra)
#else
	rpcc     R0
	zapnot   R0,0xf,R0
#endif
/*}}}  */
	bis     R0,R0,REG_0
	jmp     zero,(L)
! -------------------------------------------------------------------------
/*}}}  */
/*{{{  altwt*/
!  no parameters or results
	.globl  $altwt
$altwt:
	lda     T0,NoneSelected_o(zero)
	stl     T0,0(Wptr)              # mark not selected
	ldl     T0,W_State(Wptr)
	addl    MIN,Ready_d,T1
	cmpeq   T0,T1,P5                # Is State = Ready?
	beq     P5,1f
	jmp     zero,(L)

1:      addl    MIN,Waiting_d,T1
	stl     T1,W_State(Wptr)        # State := Waiting
	br      $stopp
! -------------------------------------------------------------------------

/*}}}  */
/*{{{  disc*/
!  on entry: REG_2 (Transputer A) is process address
!            REG_1 (Transputer B) is boolean guard
!            REG_0 (Transputer C) is channel address
!  on exit:  REG_0 (Transputer A) is "selected" flag

	.globl  $disc
$disc:
	beq     REG_1,chanreturnfalse
	ldl     T0,0(REG_0)              # channel word
	cmpeq   MIN,T0,P5                # channel = empty?
	bne     P5,chanreturnfalse
	cmpeq   T0,Wptr,P5               # channel = W
	beq     P5,selectiffirst
	stl     MIN,0(REG_0)             # channel := empty
chanreturnfalse:
	bis     zero,zero,REG_0          # selected-flag := false
	jmp     zero,(L)
selectiffirst:
	ldl     T0,0(Wptr)
	lda     T1,NoneSelected_o(zero)
	cmpeq   T0,T1,P5                 # W[0] = empty?
	beq     P5,chanreturnfalse
1:      stl     REG_2,0(Wptr)            # W[0] := codeptr
	lda     REG_0,True(zero)         # selected-flag := true
	jmp     zero,(L)
! -------------------------------------------------------------------------

/*}}}  */
/*{{{  rangerr*/
! on entry P0 has trap reason (rangerr = 2)
	.globl  $rangerr
$rangerr:
	.globl  $seterr
$seterr:
     .set at
	stq      L, OccErrPtr            # address of next occam instruction
     .set noat
	bis      L,L,P1
	lda      pv,s occam_exit
	jsr      ra,(pv),occam_exit      # will not return
/*}}}  */
/*}}}  */

/* If something seriously impossible happens  - escape     */
/********************************************************************************/
imposs:
     .set at
	stq     L, OccErrPtr            # address of next occam instruction
     .set noat

	lda     pv,impossible
	jsr     ra,(pv),impossible
	.end    enter
